import shlex
from typing import Callable

from v6d3music.core.default_effects import *
from v6d3music.core.mainservice import MainService
from v6d3music.utils.assert_admin import assert_admin
from v6d3music.utils.catch import *
from v6d3music.utils.effects_for_preset import *
from v6d3music.utils.options_for_effects import *
from v6d3music.utils.presets import allowed_presets

from v6d2ctx.at_of import AtOf
from v6d2ctx.context import Context, Explicit, command_type
from v6d2ctx.lock_for import lock_for

__all__ = ('get_of',)


def get_of(mainservice: MainService) -> Callable[[str], command_type]:
    at_of: AtOf[str, command_type] = AtOf()
    at, of = at_of()

    @at('help')
    async def help_(ctx: Context, args: list[str]) -> None:
        match args:
            case []:
                await ctx.reply('music bot')
            case [name]:
                await ctx.reply(f'help for {name}: `{name} help`')

    @at('/')
    @at('play')
    async def play(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args,
            f'''
    `play ...args`
    `play url [- effects]/[+ preset] [[[h]]] [[m]] [s] [tor] ...args`
    `pause`
    `resume`
    presets: {shlex.join(allowed_presets)}
    ''',
            (), 'help'
        )
        async with lock_for(ctx.guild, 'not in a guild'):
            queue = await mainservice.context(ctx, create=True, force_play=False).queue()
            if ctx.message.attachments:
                if len(ctx.message.attachments) > 1:
                    raise Explicit('no more than one attachment')
                args = [ctx.message.attachments[0].url] + args
            async for audio in mainservice.yt_audios(ctx, args):
                queue.append(audio)
            await ctx.reply('done')

    @at('skip')
    async def skip(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `skip [first] [last]`
    ''', 'help'
        )
        assert ctx.member is not None
        match args:
            case []:
                queue = await mainservice.context(ctx, create=False, force_play=False).queue()
                queue.skip_at(0, ctx.member)
            case [pos] if pos.isdecimal():
                pos = int(pos)
                queue = await mainservice.context(ctx, create=False, force_play=False).queue()
                queue.skip_at(pos, ctx.member)
            case [pos0, pos1] if pos0.isdecimal() and pos1.isdecimal():
                pos0, pos1 = int(pos0), int(pos1)
                queue = await mainservice.context(ctx, create=False, force_play=False).queue()
                for i in range(pos0, pos1 + 1):
                    if not queue.skip_at(pos0, ctx.member):
                        pos0 += 1
            case _:
                raise Explicit('misformatted')
        await ctx.reply('done')

    @at('to')
    async def skip_to(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `to [[h]] [m] s`
    ''', 'help'
        )
        match args:
            case [h, m, s] if h.isdecimal() and m.isdecimal() and s.isdecimal():
                seconds = 3600 * int(h) + 60 * int(m) + int(s)
            case [m, s] if m.isdecimal() and s.isdecimal():
                seconds = 60 * int(m) + int(s)
            case [s] if s.isdecimal():
                seconds = int(s)
            case _:
                raise Explicit('misformatted')
        queue = await mainservice.context(ctx, create=False, force_play=False).queue()
        queue.queue[0].set_seconds(seconds)

    @at('effects')
    async def effects_(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `effects - effects`
    `effects + preset`
    ''', 'help'
        )
        match args:
            case ['-', effects]:
                pass
            case ['+', preset]:
                effects = effects_for_preset(preset)
            case _:
                raise Explicit('misformatted')
        assert_admin(ctx.member)
        queue = await mainservice.context(ctx, create=False, force_play=False).queue()
        yta = queue.queue[0]
        seconds = yta.source_seconds()
        yta.options = options_for_effects(effects)
        yta.set_seconds(seconds)

    @at('default')
    async def default(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `default - effects`
    `default + preset`
    `default none`
    ''', 'help'
        )
        assert ctx.guild is not None
        match args:
            case ['-', effects]:
                pass
            case ['+', preset]:
                effects = effects_for_preset(preset)
            case ['none']:
                effects = None
            case []:
                await ctx.reply(f'current default effects: {mainservice.defaulteffects.get(ctx.guild.id)}')
                return
            case _:
                raise Explicit('misformatted')
        assert_admin(ctx.member)
        await mainservice.defaulteffects.set(ctx.guild.id, effects)
        await ctx.reply(f'effects set to `{effects}`')

    @at('repeat')
    async def repeat(ctx: Context, args: list[str]):
        match args:
            case []:
                n = 1
            case [n_] if n_.isdecimal():
                n = int(n_)
            case _:
                raise Explicit('misformatted')
        assert_admin(ctx.member)
        queue = await mainservice.context(ctx, create=False, force_play=False).queue()
        queue.repeat(n)

    @at('shuffle')
    async def shuffle(ctx: Context, args: list[str]):
        assert_admin(ctx.member)
        queue = await mainservice.context(ctx, create=False, force_play=False).queue()
        queue.shuffle()

    @at('branch')
    async def branch(ctx: Context, args: list[str]):
        match args:
            case ['-', effects]:
                pass
            case ['+', preset]:
                effects = effects_for_preset(preset)
            case ['none']:
                effects = ''
            case []:
                effects = None
            case _:
                raise Explicit('misformatted')
        assert_admin(ctx.member)
        queue = await mainservice.context(ctx, create=False, force_play=False).queue()
        if not queue.queue:
            raise Explicit('empty queue')
        audio = queue.queue[0].branch()
        if effects is not None:
            seconds = audio.source_seconds()
            audio.options = options_for_effects(effects or None)
            audio.set_seconds(seconds)
        else:
            audio.set_source()
        queue.queue.insert(1, audio)

    @at('//')
    @at('queue')
    async def queue_(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `queue`
    `queue clear`
    `queue resume`
    `queue pause`
    ''', 'help'
        )
        assert ctx.member is not None
        match args:
            case []:
                await ctx.long(
                    (
                        await (
                            await mainservice.context(ctx, create=True, force_play=False).queue()
                        ).format()
                    ).strip() or 'no queue'
                )
            case ['clear']:
                (await mainservice.context(ctx, create=False, force_play=False).queue()).clear(ctx.member)
                await ctx.reply('done')
            case ['resume']:
                async with lock_for(ctx.guild, 'not in a guild'):
                    await mainservice.context(ctx, create=True, force_play=True).vc()
                    await ctx.reply('done')
            case ['pause']:
                async with lock_for(ctx.guild, 'not in a guild'):
                    vc = await mainservice.context(ctx, create=True, force_play=False).vc()
                    vc.pause()
                    await ctx.reply('done')
            case _:
                raise Explicit('misformatted')

    @at('swap')
    async def swap(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `swap a b`
    ''', 'help'
        )
        assert ctx.member is not None
        match args:
            case [a, b] if a.isdecimal() and b.isdecimal():
                a, b = int(a), int(b)
                (await mainservice.context(ctx, create=False, force_play=False).queue()).swap(ctx.member, a, b)
            case _:
                raise Explicit('misformatted')

    @at('move')
    async def move(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `move a b`
    ''', 'help'
        )
        assert ctx.member is not None
        match args:
            case [a, b] if a.isdecimal() and b.isdecimal():
                a, b = int(a), int(b)
                (await mainservice.context(ctx, create=False, force_play=False).queue()).move(ctx.member, a, b)
            case _:
                raise Explicit('misformatted')

    @at('volume')
    async def volume_(ctx: Context, args: list[str]) -> None:
        await catch(
            ctx, args, '''
    `volume volume`
    ''', 'help'
        )
        assert ctx.member is not None
        match args:
            case [volume]:
                volume = float(volume)
                await (await mainservice.context(ctx, create=True, force_play=False).main()).set(volume, ctx.member)
            case _:
                raise Explicit('misformatted')

    @at('pause')
    async def pause(ctx: Context, _args: list[str]) -> None:
        vc = await mainservice.context(ctx, create=False, force_play=False).vc()
        vc.pause()

    @at('resume')
    async def resume(ctx: Context, _args: list[str]) -> None:
        vc = await mainservice.context(ctx, create=False, force_play=True).vc()
        vc.resume()

    return of
