import asyncio
import contextlib
import os
import sys
import time

import discord
from v6d3music.app import AppContext
from v6d3music.commands import *
from v6d3music.config import prefix
from v6d3music.core.caching import *
from v6d3music.core.default_effects import *
from v6d3music.core.mainservice import MainService

from ptvp35 import *
from rainbowadn.instrument import Instrumentation
from v6d1tokens.client import request_token
from v6d2ctx.handle_content import handle_content
from v6d2ctx.pain import ABlockMonitor
from v6d2ctx.serve import serve

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)


class MusicClient(discord.Client):
    pass


client = MusicClient(
    intents=discord.Intents(
        members=True,
        guilds=True,
        bans=True,
        emojis=True,
        invites=True,
        voice_states=True,
        guild_messages=True,
        reactions=True,
        message_content=True,
    ),
    loop=loop,
)


banned_guilds = set(map(int, map(str.strip, os.getenv('banned_guilds', '').split(':'))))


def guild_allowed(guild: discord.Guild | None) -> bool:
    return guild is not None and guild.id not in banned_guilds


def message_allowed(message: discord.Message) -> bool:
    return guild_allowed(message.guild)


def register_handlers(mainservice: MainService):
    of = get_of(mainservice)

    @client.event
    async def on_message(message: discord.Message) -> None:
        if message_allowed(message):
            await handle_content(of, message, message.content, prefix, client)

    @client.event
    async def on_ready():
        print('ready')
        await client.change_presence(
            activity=discord.Game(
                name='феноменально',
            )
        )
        await mainservice.restore()


class UpgradeABMInit(Instrumentation):
    def __init__(self):
        super().__init__(ABlockMonitor, '__init__')

    def instrument(self, method, abm, *, threshold=0.0, delta=10.0, interval=0.0):
        print('created upgraded')
        method(abm, threshold=threshold, delta=delta, interval=interval)
        abm.threshold = threshold


class UpgradeABMTask(Instrumentation):
    def __init__(self):
        super().__init__(ABlockMonitor, '_monitor')

    async def instrument(self, _, abm):
        print('started upgraded')
        while True:
            delta = abm.delta
            t = time.time()
            await asyncio.sleep(delta)
            spent = time.time() - t
            delay = spent - delta
            if delay > abm.threshold:
                abm.threshold = delay
                print(
                    f'upgraded block monitor reached new peak delay {delay:.4f}')
            interval = abm.interval
            if interval > 0:
                await asyncio.sleep(interval)


def _upgrade_abm() -> contextlib.ExitStack:
    with contextlib.ExitStack() as es:
        es.enter_context(UpgradeABMInit())
        es.enter_context(UpgradeABMTask())
        return es.pop_all()
    raise RuntimeError


class PathPrint(Instrumentation):
    def __init__(self, methodname: str, pref: str):
        super().__init__(DbConnection, methodname)
        self.pref = pref

    async def instrument(self, method, db: DbConnection, *args, **kwargs):
        result = await method(db, *args, **kwargs)
        try:
            print(self.pref, db._DbConnection__path)  # type: ignore
        except Exception:
            from traceback import print_exc
            print_exc()
        return result


def _db_ee() -> contextlib.ExitStack:
    with contextlib.ExitStack() as es:
        es.enter_context(PathPrint('_initialize', 'open :'))
        es.enter_context(PathPrint('aclose', 'close:'))
        return es.pop_all()
    raise RuntimeError


async def main():
    async with (
        DefaultEffects() as defaulteffects,
        MainService(defaulteffects, client) as mainservice,
        AppContext(mainservice),
        ABlockMonitor(delta=0.5)
    ):
        register_handlers(mainservice)
        if 'guerilla' in sys.argv:
            from pathlib import Path
            tokenpath = Path('.token.txt')
            if tokenpath.exists():
                token = tokenpath.read_text()
            else:
                token = input('token:')
                tokenpath.write_text(token)
        elif (token_ := os.getenv('trial_token')):
            token = token_
        else:
            token = await request_token('music', 'token')
        await client.login(token)
        if os.getenv('v6tor', None) is None:
            print('no tor')
        await client.connect()
        print('exited')


if __name__ == '__main__':
    with _upgrade_abm(), _db_ee():
        serve(main(), client, loop)
