from typing import Any, AsyncIterable, Callable

from v6d3music.utils.effects_for_preset import effects_for_preset
from v6d3music.utils.entries_for_url import entries_for_url
from v6d3music.utils.options_for_effects import options_for_effects
from v6d3music.utils.sparq import sparq
from v6d2ctx.context import Explicit

__all__ = ('InfoCtx', 'UrlCtx', 'ArgCtx',)


class InfoCtx:
    def __init__(
        self, info: dict[str, Any], effects: str | None, already_read: int, tor: bool, ignore: bool
    ) -> None:
        self.info = info
        self.effects = effects
        self.already_read = already_read
        self.tor = tor
        self.ignore = ignore


class UrlCtx:
    def __init__(self, url: str, effects: str | None) -> None:
        self.url = url
        self.effects = effects
        self.already_read = 0
        self.tor = False
        self.ignore = False

    async def entries(self) -> AsyncIterable[InfoCtx]:
        try:
            async for info in entries_for_url(self.url, self.tor):
                yield InfoCtx(info, self.effects, self.already_read, self.tor, self.ignore)
        except Exception:
            if not self.ignore:
                raise


class ArgCtx:
    def __init__(self, default_effects: str | None, args: list[str]) -> None:
        self.sources: list[UrlCtx] = []
        while args:
            match args:
                case [url, '-', effects, *args]:
                    pass
                case [url, '+', preset, *args]:
                    effects = effects_for_preset(preset)
                case [url, *args]:
                    effects = default_effects
                case _:
                    raise RuntimeError
            ctx = UrlCtx(url, effects)
            seconds = 0
            match args:
                case [h, m, s, *args] if h.isdecimal() and m.isdecimal() and s.isdecimal():
                    seconds = 3600 * int(h) + 60 * int(m) + int(s)
                case [m, s, *args] if m.isdecimal() and s.isdecimal():
                    seconds = 60 * int(m) + int(s)
                case [s, *args] if s.isdecimal():
                    seconds = int(s)
                case [*args]:
                    pass
            ctx.already_read = round(seconds / sparq(options_for_effects(effects)))
            while True:
                match args:
                    case ['tor', *args]:
                        if ctx.tor:
                            raise Explicit('duplicate tor')
                        ctx.tor = True
                    case ['ignore', *args]:
                        if ctx.ignore:
                            raise Explicit('duplicate ignore')
                        ctx.ignore = True
                    case [*args]:
                        break
            self.sources.append(ctx)
